"""Work with environment variables in a GitHub Actions workflow."""

from typing import Type as _Type
import os as _os
from pathlib import Path as _Path
import json as _json

from actionman import _format
from actionman.exception import (
    ActionManGitHubError as _ActionManGitHubError,
    ActionManInputVariableDeserializationError as _ActionManInputVariableDeserializationError,
    ActionManInputVariableTypeError as _ActionManInputVariableTypeError,
    ActionManInputVariableTypeMismatchError as _ActionManInputVariableTypeMismatchError,
)


_ENV_VAR_NAME = "GITHUB_ENV"
_FILEPATH: _Path | None = _Path(_os.environ[_ENV_VAR_NAME]) if _ENV_VAR_NAME in _os.environ else None


def read(
    name: str,
    typ: _Type[str | bool | int | float | list | dict] = str,
) -> str | bool | int | float | list | dict | None:
    """Read an environment variable and cast it to the given type.

    Parameters
    ----------
    name : str
        The name of the environment variable to read.
    typ : Type[str | bool | int | float | list | dict], default: str
        The type to cast the environment variable to.
        If the type is not str, the value of the environment variable
        is expected to be a JSON string that can be deserialized to the given type.

    Returns
    -------
    str | bool | int | float | list | dict | None
        The value of the environment variable cast to the given type.
        If the environment variable is not set, None is returned.

    Raises
    ------
    actionman.exception.ActionManInputVariableTypeError
        If the specified type is not supported.
    actionman.exception.ActionManInputVariableDeserializationError
        If the environment variable could not be deserialized.
    actionman.exception.ActionManInputVariableTypeMismatchError
        If the deserialized environment variable has a type other than the specified type.
    """
    if typ not in (str, bool, int, float, list, dict):
        raise _ActionManInputVariableTypeError(var_name=name, var_type=typ)
    value = _os.environ.get(name)
    if value is None:
        return
    if typ is str:
        return value
    try:
        value_deserialized = _json.loads(value)
    except Exception as e:
        raise _ActionManInputVariableDeserializationError(
            var_name=name,
            var_value=value,
            var_type=typ,
            exception=e
        ) from e
    if not isinstance(value_deserialized, typ):
        raise _ActionManInputVariableTypeMismatchError(
            var_name=name,
            var_value=value_deserialized,
            var_type=typ
        )
    return value_deserialized


def write(name: str, value: dict | list | tuple | str | bool | int | float | None) -> str:
    """Set a persistent environment variable
    that is available to all subsequent steps in the current job.

    This is done by writing the environment variable
    to the environment file whose path is specified by the 'GITHUB_ENV' environment variable.

    Parameters
    ----------
    name : str
        The name of the output parameter.
    value : dict | list | tuple | str | bool | int | float | None
        The value of the output parameter.
        If the value is not a string, it will be serialized and written as a JSON string.

    Returns
    -------
    str
        The output that was written to the file.
        This is only useful for logging/debugging purposes.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_ENV' environment variable is not set.
    actionman.exception.ActionManOutputVariableTypeError
        If the value has an unsupported type.
    actionman.exception.ActionManOutputVariableSerializationError
        If the value could not be serialized to a JSON string.

    References
    ----------
    - [GitHub Docs: Workflow Commands for GitHub Actions: Setting an environment variable](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-environment-variable)
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    output = _format.output_variable(key=name, value=value)
    with open(_FILEPATH, "a") as f:
        print(output, file=f)
    return output



"""Create workflow annotations and logs for a GitHub Actions workflow run."""

from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING

from rich.console import Console, Group as _Group
from rich import segment as _segment
from rich.text import Text as _Text

if _TYPE_CHECKING:
    from typing import Literal
    from rich.console import RenderableType
    from rich.text import TextType
    from protocolman import Stringable


class Logger:

    def __init__(self, console: Console | dict | None = None):
        if not console:
            self.console = Console(
                color_system="truecolor",
                force_terminal=True,
                emoji_variant="emoji",
                width=90,
                safe_box=False,
            )
        elif isinstance(console, dict):
            self.console = Console(**console)
        else:
            self.console = console
        return


    def debug(self, *contents: RenderableType, out: bool = True) -> _Group:
        """Create a debug log.

        Parameters
        ----------
        contents : actionman.protocol.Stringable
            The log message.
        out : bool, default: True
            Whether to directly print the debug log.

        Returns
        -------
        str
            The debug log.

        References
        ----------
        - [GitHub Docs: Workflow Commands for GitHub Actions: Setting a debug message](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-a-debug-message)
        """
        final_lines: list[_segment.Segments] = []
        for content in contents:
            for line in self.console.render_lines(content, new_lines=True):
                line.insert(0, _segment.Segment("::debug::"))
                final_lines.append(_segment.Segments(line))
        output = _Group(*final_lines, fit=False)
        if out:
            self.console.print(output)
        return output


    def annotation(
        self,
        typ: Literal["notice", "warning", "error"],
        message: TextType,
        title: TextType = "",
        filename: Stringable = "",
        line_start: int = 0,
        line_end: int = 0,
        column_start: int = 0,
        column_end: int = 0,
        out: bool = True,
    ) -> _Text:
        """Create a notice, warning, or error annotation.

        Parameters
        ----------
        typ : {"notice", "warning", "error"}
            The type of annotation to create.
        message : actionman.protocol.Stringable
            The annotation message.
        title : actionman.protocol.Stringable, optional
            The annotation title.
        filename : actionman.protocol.Stringable, optional
            Path to a file in the repository to associate the message with.
        line_start : int, optional
            The starting line number in the file specified by the 'filename' argument,
            to associate the message with.
        line_end : int, optional
            The ending line number in the file specified by the 'filename' argument,
            to associate the message with.
        column_start : int, optional
            The starting column number in the line specified by the 'line_start' argument,
            to associate the message with.
        column_end : int, optional
            The ending column number in the line specified by the 'line_start' argument,
            to associate the message with.
        out : bool, default: True
            Whether to directly print the annotation.

        Returns
        -------
        str
            The annotation.

        References
        ----------
        - [GitHub Docs: Workflow Commands for GitHub Actions: Setting a notice message](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-a-notice-message)
        - [GitHub Docs: Workflow Commands for GitHub Actions: Setting a warning message](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-a-warning-message)
        - [GitHub Docs: Workflow Commands for GitHub Actions: Setting an error message](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-error-message)
        """
        args = locals()
        args_str = []
        for arg_name, github_arg_name in (
            ("title", "title"),
            ("filename", "file"),
            ("line_start", "line"),
            ("line_end", "endLine"),
            ("column_start", "col"),
            ("column_end", "endColumn"),
        ):
            if args[arg_name]:
                args_str.append(f"{github_arg_name}={args[arg_name]}")
        args_str_full = ",".join(args_str)
        sig_section = f" {args_str_full}" if args_str_full else ""
        output = _Text(f"::{typ}{sig_section}::")
        output.append(message)
        if out:
            self.console.print(output)
        return output


    def group(self, *contents: RenderableType, title: TextType | None = None, out: bool = True) -> _Group:
        """Create an expandable log group.

        Parameters
        ----------
        contents : actionman.protocol.Stringable
            The details of the log group.
        title : actionman.protocol.Stringable
            The title of the log group.
        out : bool, default: True
            Whether to directly print the log group.

        Returns
        -------
        str
            The log group.

        References
        ----------
        - [GitHub Docs: Workflow Commands for GitHub Actions: Grouping log output](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#grouping-log-output)
        """
        start = self.group_open(title, out=False)
        end = self.group_close(out=False)
        output = _Group(start, *contents, end)
        if out:
            self.console.print(output)
        return output


    def group_open(self, title: TextType | None = None, out: bool = True) -> _Text:
        """Open an expandable log group.

        Parameters
        ----------
        title : actionman.protocol.Stringable
            The title of the log group.
        out : bool, default: True
            Whether to directly print the log group.

        Returns
        -------
        str
            The log group's opening tag.

        References
        ----------
        - [GitHub Docs: Workflow Commands for GitHub Actions: Grouping log output](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#grouping-log-output)
        """
        output = _Text("::group::")
        if title:
            output.append(title)
        if out:
            self.console.print(output)
        return output


    def group_close(self, out: bool = True) -> str:
        """Close an expandable log group.

        Parameters
        ----------
        out
            Whether to directly print the log group.

        Returns
        -------
        The log group's closing tag.

        References
        ----------
        - [GitHub Docs: Workflow Commands for GitHub Actions: Grouping log output](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#grouping-log-output)
        """
        output = "::endgroup::"
        if out:
            self.console.print(output)
        return output



"""Work with the step summary of the current step of a job run in a GitHub Actions workflow.

References
----------
- [GitHub Docs: Workflow Commands for GitHub Actions: Adding a job summary](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#adding-a-job-summary)
"""

from __future__ import annotations as _annotations
from typing import TYPE_CHECKING as _TYPE_CHECKING
import os as _os
from pathlib import Path as _Path

from actionman.exception import ActionManGitHubError as _ActionManGitHubError

if _TYPE_CHECKING:
    from protocolman import Stringable


_ENV_VAR_NAME = "GITHUB_STEP_SUMMARY"
_FILEPATH: _Path | None = _Path(_os.environ[_ENV_VAR_NAME]) if _ENV_VAR_NAME in _os.environ else None


def filepath() -> _Path:
    """Get the path to the file where the step summary is stored.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_STEP_SUMMARY' environment variable is not set.
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    return _FILEPATH


def read() -> str | None:
    """Read the current step summary contents from the file.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_STEP_SUMMARY' environment variable is not set.
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    return _FILEPATH.read_text() if _FILEPATH.is_file() else None


def append(content: Stringable) -> None:
    """Append the given content to the step summary file.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_STEP_SUMMARY' environment variable is not set.
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    with open(_FILEPATH, "a") as f:
        print(content, file=f)
    return


def write(content: Stringable) -> None:
    """Overwrite the step summary file with the given content.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_STEP_SUMMARY' environment variable is not set.
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    with open(_FILEPATH, "w") as f:
        print(content, file=f)
    return


def remove() -> None:
    """Remove all step summary contents by deleting the file.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_STEP_SUMMARY' environment variable is not set.
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    _FILEPATH.unlink()
    return



from types import NoneType as _NoneType
import base64 as _base64
import json as _json

from actionman.exception import (
    ActionManOutputVariableTypeError as _ActionManOutputVariableTypeError,
    ActionManOutputVariableSerializationError as _ActionManOutputVariableSerializationError,
)


def output_variable(key: str, value: dict | list | tuple | str | bool | int | float | None) -> str:
    """Format a key-value pair for output.

    This converts a variable name and its corresponding value to a string
    that could be written as an environment variable (to `GITHUB_ENV`)
    or output parameter (to `GITHUB_OUTPUT`) in a job step of a GitHub Actions workflow.

    Parameters
    ----------
    key
        Name of the variable.
    value
        Value of the variable.

    Raises
    ------
    Actionman.exception.ActionManOutputVariableTypeError
        If the value has an unsupported type.
    Actionman.exception.ActionManOutputVariableSerializationError
        If the value could not be serialized to a JSON string.

    References
    ----------
    - [GitHub Docs: Workflow commands for GitHub Actions: Setting an environment variable](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-environment-variable)
    - [GitHub Docs: Workflow commands for GitHub Actions: Setting an output parameter](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-output-parameter)
    """
    if isinstance(value, str):
        if "\n" in value:
            with open("/dev/urandom", "rb") as f:
                random_bytes = f.read(15)
            random_delimeter = _base64.b64encode(random_bytes).decode("utf-8")
            return f"{key}<<{random_delimeter}\n{value}\n{random_delimeter}"
    elif isinstance(value, (dict, list, tuple, bool, int, float, _NoneType)):
        try:
            value = _json.dumps(value)
        except Exception as e:
            raise _ActionManOutputVariableSerializationError(
                var_name=key,
                var_value=value,
                exception=e,
            ) from e
    else:
        raise _ActionManOutputVariableTypeError(
            var_name=key,
            var_value=value,
        )
    return f"{key}={value}"



"""Work with the step output of the current step of a job run in a GitHub Actions workflow.

References
----------
- [GitHub Docs: Workflow Commands for GitHub Actions: Setting an output parameter](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-output-parameter)
"""

import os as _os
from pathlib import Path as _Path

from actionman import _format
from actionman.exception import ActionManGitHubError as _ActionManGitHubError


_ENV_VAR_NAME = "GITHUB_OUTPUT"
_FILEPATH: _Path | None = _Path(_os.environ[_ENV_VAR_NAME]) if _ENV_VAR_NAME in _os.environ else None


def write(name: str, value: dict | list | tuple | str | bool | int | float | None) -> str:
    """Set an output parameter for the current step.

    This is done by writing the output
    to the environment file whose path is specified by the 'GITHUB_OUTPUT' environment variable.

    Parameters
    ----------
    name : str
        The name of the output parameter.
    value : dict | list | tuple | str | bool | int | float | None
        The value of the output parameter.
        If the value is not a string, it will be serialized and written as a JSON string.

    Returns
    -------
    str
        The output that was written to the file.
        This is only useful for logging/debugging purposes.

    Raises
    ------
    actionman.exception.ActionManGitHubError
        If the 'GITHUB_OUTPUT' environment variable is not set.
    actionman.exception.ActionManOutputVariableTypeError
        If the value has an unsupported type.
    actionman.exception.ActionManOutputVariableSerializationError
        If the value could not be serialized to a JSON string.
    """
    if not _FILEPATH:
        raise _ActionManGitHubError(missing_env_var=_ENV_VAR_NAME)
    output = _format.output_variable(key=name, value=value)
    with open(_FILEPATH, "a") as f:
        print(output, file=f)
    return output



"""ActionMan: GitHub Actions utilities for Python."""

import io as _io
import sys as _sys

from actionman import log, env_var, step_output, step_summary


def in_gha() -> bool:
    """Check whether the current program is running in a GitHub Actions environment.

    This is done by checking the presence of the 'GITHUB_ACTIONS' environment variable.

    References
    ----------
    - [GitHub Docs: Default environment variables](https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/store-information-in-variables#default-environment-variables)
    """
    return bool(env_var.read("GITHUB_ACTIONS", bool))


if hasattr(_sys.stdout, 'buffer'):
    # Wrap the standard output stream to change its encoding to UTF-8,
    # which is required for writing unicode characters (e.g., emojis) to the console in Windows.
    # However, this works in standard Python environments where sys.stdout is a regular file object;
    # in environments like Jupyter, sys.stdout is already set up to handle Unicode,
    # and does not need to be (and cannot be) wrapped in this way.
    _sys.stdout = _io.TextIOWrapper(_sys.stdout.buffer, encoding='utf-8')



"""Exceptions raised by ActionMan."""


from typing import Any as _Any, Type as _Type


class ActionManError(Exception):
    """Base class for all ActionMan errors."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)
        return


class ActionManGitHubError(ActionManError):
    """The package was initialized from outside GitHub, and the required environment variables are not set."""

    def __init__(self, missing_env_var: str):
        message = (
            f"The package was initialized from outside GitHub, "
            f"as the required environment variable '{missing_env_var}' is not set."
        )
        super().__init__(message)
        self.missing_environment_variable: str = missing_env_var
        return


class ActionManVariableError(ActionManError):
    """Base class for all ActionMan variable errors."""

    def __init__(self, var_name: _Any, message: str):
        super().__init__(message)
        self.variable_name = var_name
        return


class ActionManInputVariableError(ActionManVariableError):
    """Base class for all ActionMan input variable errors."""

    def __init__(self, var_name: str, message: str):
        super().__init__(var_name=var_name, message=message)
        return


class ActionManOutputVariableError(ActionManVariableError):
    """Base class for all ActionMan output variable errors."""

    def __init__(self, var_name: str, var_value: _Any, message: str):
        super().__init__(var_name=var_name, message=message)
        self.variable_value = var_value
        return


class ActionManInputVariableTypeError(ActionManInputVariableError):
    """An input variable has an unsupported type."""

    def __init__(self, var_name: str, var_type: _Any):
        message = (
            f"Input variable '{var_name}' has an unsupported expected type '{var_type.__name__}'."
        )
        super().__init__(var_name=var_name, message=message)
        return


class ActionManInputVariableTypeMismatchError(ActionManInputVariableError):
    """An input variable has a type other than its expected type."""

    def __init__(self, var_name: str, var_value: str | bool | int | float | list | dict, var_type: _Any):
        message = (
            f"Input variable '{var_name}' has type '{type(var_value).__name__}', "
            f"but expected type {var_type.__name__}."
        )
        super().__init__(var_name=var_name, message=message)
        self.variable_value = var_value
        self.variable_type = var_type
        return


class ActionManInputVariableDeserializationError(ActionManInputVariableError):
    """An input variable could not be deserialized from a JSON string."""

    def __init__(
        self,
        var_name: str,
        var_value: str,
        var_type: _Type[bool | int | float | list | dict],
        exception: Exception
    ):
        message = (
            f"Failed to deserialize input variable '{var_name}' with expected type '{var_type.__name__}' "
            f"from a JSON string; {str(exception).removesuffix('.')}."
        )
        super().__init__(var_name=var_name, message=message)
        self.variable_value = var_value
        self.variable_type = var_type
        return


class ActionManOutputVariableSerializationError(ActionManOutputVariableError):
    """An output variable could not be serialized to a JSON string."""

    def __init__(self, var_name: str, var_value: _Any, exception: Exception):
        message = (
            f"Failed to serialize output variable '{var_name}' with type '{type(var_value).__name__}' "
            f"to a JSON string; {str(exception).removesuffix('.')}."
        )
        super().__init__(var_name=var_name, var_value=var_value, message=message)
        return


class ActionManOutputVariableTypeError(ActionManOutputVariableError):
    """An output variable has an unsupported type."""

    def __init__(self, var_name: str, var_value: _Any):
        message = (
            f"Output variable '{var_name}' has an unsupported type '{type(var_value).__name__}'."
        )
        super().__init__(var_name=var_name, var_value=var_value, message=message)
        return
